package edu.princeton.swing;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.Vector;
import java.util.Iterator;

import javax.swing.*;
import javax.swing.text.*;

/**
 * PTextField extends JTextField to provide ClipboardTarget support.
 *
 * @author btsang
 * @version 7.1
 */

public class PTextField extends JTextField implements ClipboardTarget
{
    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    private JPopupMenu popupMenu;
    private Vector actionListeners;
    private String actionCommand;

    /**
     * Constructs a new PTextField.
     */
    public PTextField()
    {
        super();

        initTextField();
    }

    /**
     * Constructs a new PTextField.
     */
    public PTextField(Document doc, String text, int columns)
    {
        super(doc, text, columns);

        initTextField();
    }

    /**
     * Constructs a new PTextField.
     */
    public PTextField(int columns)
    {
        super(columns);

        initTextField();
    }

    /**
     * Constructs a new PTextField.
     */
    public PTextField(String text)
    {
        super(text);

        initTextField();
    }

    /**
     * Constructs a new PTextField.
     */
    public PTextField(String text, int columns)
    {
        super(text, columns);

        initTextField();
    }

    /**
     * Called by the constructors to init the TList properly.
     */
    private void initTextField()
    {
        popupMenu = ClipboardTargetManager.createDefaultPopupMenu();
        actionListeners = new Vector();
        actionCommand = null;

        ClipboardTargetManager.installMappings(this);

        enableEvents(AWTEvent.FOCUS_EVENT_MASK | AWTEvent.KEY_EVENT_MASK |
                     AWTEvent.MOUSE_EVENT_MASK);
    }

    /**
     * Adds a listener to monitor action events (enter key events) on the PTextField.
     *
     * @param listener The listener to add to this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void addActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.add(listener);
    }

    /**
     * Removes a listener from this component.  Nothing will happen if no match was found.
     *
     * @param listener The listener to remove from this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void removeActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.remove(listener);
    }

    /**
     * Returns the action command associated with the enter key being pressed.  A null value
     * indicates thatthe text of the text field at the time of the event will be used.
     *
     * return The action command to be associated with the enter key.
     */
    public String getActionCommand()
    {
        return actionCommand;
    }

    /**
     * Sets the action command associated with the enter key being pressed.
     *
     * @param actionCommand The action command to be associated with the enter key.  If null is
     * passed, the text of the text field at the time of the event will be used.
     */
    public void setActionCommand(String actionCommand)
    {
        if (this.actionCommand != actionCommand && 
            (actionCommand != null || !actionCommand.equals(this.actionCommand)))
        {
            String oldValue = this.actionCommand;
            this.actionCommand = actionCommand;

            firePropertyChange("actionCommand", oldValue, actionCommand);
        }
    }

    /**
     * Returns wheter or not the cut operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to cut() should be permitted.
     */
    public boolean canCut()
    {
        return isEnabled() && isEditable();
    }

    /**
     * Returns wheter or not the copy operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to copy() should be permitted.
     */
    public boolean canCopy()
    {
        return isEnabled();
    }

    /**
     * Returns wheter or not the paste operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to paste() should be permitted.
     */
    public boolean canPaste()
    {
        return isEnabled() && isEditable();
    }

    /**
     * Returns wheter or not the selectAll operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to selectAll() should be permitted.
     */
    public boolean canSelectAll()
    {
        return isEnabled();
    }

    /**
     * Intercept KeyEvents to send off ActionEvents when enter is pressed.
     */
    protected void processKeyEvent(KeyEvent e)
    {
        if (e.getKeyCode() == KeyEvent.VK_ENTER)
        {
            // Dispatch the events
            if (e.getID() == KeyEvent.KEY_PRESSED && !actionListeners.isEmpty())
            {
                ActionEvent actionEvent;
                String actionCommand = this.actionCommand;
                if (actionCommand == null)
                    actionCommand = getText();
                
                actionEvent = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, actionCommand);

                Object array[] = actionListeners.toArray();
                for (int ctr = 0; ctr < array.length; ctr++)
                    ((ActionListener)array[ctr]).actionPerformed(actionEvent);
            }

            e.consume();
        }

        super.processKeyEvent(e);
    }

    /**
     * Intercept FocusEvents to notify the ClipboardTargetManager of the event.
     */
    protected void processFocusEvent(FocusEvent e)
    {
        if (e.getID() == FocusEvent.FOCUS_GAINED)
        {
            ClipboardTargetManager.targetGainedFocus(this);
        }
        else if (e.getID() == FocusEvent.FOCUS_LOST)
        {
            ClipboardTargetManager.targetLostFocus(this);
        }

        super.processFocusEvent(e);
    }

    /**
     * Intercept MouseEvents to activate the popupMenu at the right time.
     */
    protected void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        boolean passedToSuper = false;

        if (popupMenu != null && e.isPopupTrigger())
        {
            if (!hasFocus())
                requestFocus();

            int x = e.getX();
            int y = e.getY();
            Point componentLocation = getLocationOnScreen();
            Dimension screenSize = TOOLKIT.getScreenSize();
            Dimension preferredSize = popupMenu.getPreferredSize();

            if (componentLocation.x + x + preferredSize.width > screenSize.width)
                x -= preferredSize.width;

            if (componentLocation.y + y + preferredSize.height > screenSize.height)
                y -= preferredSize.height;

            e.consume();
            passedToSuper = true;

            super.processMouseEvent(e);

            popupMenu.show(this, x, y);
        }

        if (!passedToSuper)
            super.processMouseEvent(e);
    }

    /**
     * Override the setEnabled() method to intercept any changes concerning the enabled status of
     * the component.
     */
    public void setEnabled(boolean enabled)
    {
        super.setEnabled(enabled);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Override the setEditable() method to intercept any changes concerning the editable status of
     * the component.
     */
    public void setEditable(boolean editable)
    {
        super.setEditable(editable);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Gets the popup menu for this component.
     *
     * @return The popup menu for the component.  Null is returned if this component does not have
     * a popup menu.
     */
    public JPopupMenu getPopupMenu()
    {
        return popupMenu;
    }

    /**
     * Sets the popup menu for this component.
     *
     * @param popupMenu The popup menu for the component.  A null value indicates that this
     * component should not have a popup menu.
     */
    public void setPopupMenu(JPopupMenu popupMenu)
    {
        if (popupMenu != this.popupMenu)
        {
            if (popupMenu != null && this.popupMenu.isVisible())
            {
                this.popupMenu.setVisible(false);
            }

            JPopupMenu oldValue = this.popupMenu;
            this.popupMenu = popupMenu;

            firePropertyChange("popupMenu", oldValue, popupMenu);
        }
    }
}
