package edu.princeton.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.beans.*;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.*;
import javax.swing.tree.*;

/**
 * PTree extends JTree to provide ClipboardTarget support.
 *
 * @author btsang
 * @version 7.1
 */

public class PTree extends JTree implements ClipboardTarget
{
    private static final Toolkit TOOLKIT = Toolkit.getDefaultToolkit();

    private JPopupMenu popupMenu;
    private Vector actionListeners;
    private String actionCommand;
    private Listener listener;

    /**
     * Constructs a new PTree.
     */
    public PTree()
    {
        super();

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(Hashtable value)
    {
        super(value);

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(Object[] value)
    {
        super(value);

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(TreeModel newModel)
    {
        super(newModel);

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(TreeNode root)
    {
        super(root);

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(TreeNode root, boolean asksAllowsChildren)
    {
        super(root, asksAllowsChildren);

        initTree();
    }

    /**
     * Constructs a new PTree.
     */
    public PTree(Vector value)
    {
        super(value);

        initTree();
    }

    /**
     * Called by the constructors to init the PTree properly.
     */
    private void initTree()
    {
        listener = new Listener();
        getSelectionModel().addPropertyChangeListener(listener);

        popupMenu = ClipboardTargetManager.createDefaultPopupMenu();

        actionListeners = new Vector();
        actionCommand = null;

        ClipboardTargetManager.installMappings(this);

        enableEvents(AWTEvent.FOCUS_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK);
    }

    /**
     * Adds a listener to monitor action events (enter key events) on the PTextField.
     *
     * @param listener The listener to add to this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void addActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.add(listener);
    }

    /**
     * Removes a listener from this component.  Nothing will happen if no match was found.
     *
     * @param listener The listener to remove from this document.  A null value will cause a
     * NullPointerException.
     */
    public synchronized void removeActionListener(ActionListener listener)
    {
        if (listener == null)
            throw new NullPointerException();

        actionListeners.remove(listener);
    }

    /**
     * Returns the action command associated with the enter key being pressed.  A null value
     * indicates thatthe text of the text field at the time of the event will be used.
     *
     * return The action command to be associated with the enter key.
     */
    public String getActionCommand()
    {
        return actionCommand;
    }

    /**
     * Sets the action command associated with the enter key being pressed.
     *
     * @param actionCommand The action command to be associated with the enter key.  If null is
     * passed, the text of the text field at the time of the event will be used.
     */
    public void setActionCommand(String actionCommand)
    {
        if (this.actionCommand != actionCommand && 
            (actionCommand != null || !actionCommand.equals(this.actionCommand)))
        {
            String oldValue = this.actionCommand;
            this.actionCommand = actionCommand;

            firePropertyChange("actionCommand", oldValue, actionCommand);
        }
    }

    /**
     * Returns wheter or not the cut operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to cut() should be permitted.
     */
    public boolean canCut()
    {
        return false;
    }

    /**
     * Copies the selection to the clipboard and remove it from the object.  The clipboard should
     * not be modified if nothing is selected.
     */
    public void cut()
    {

    }

    /**
     * Returns wheter or not the copy operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to copy() should be permitted.
     */
    public boolean canCopy()
    {
        return isEnabled();
    }

    /**
     * Copies the selection to the clipboard.  The clipboard should not be modified if nothing is
     * selected.
     */
    public void copy()
    {
        if (isEnabled())
        {
            TreePath paths[] = getSelectionPaths();
            int length = paths.length;
            if (length == 0)
                return;

            StringBuffer buffer = new StringBuffer();

            for (int ctr = 0; ctr < length; ctr++)
            {
                buffer.append(paths[ctr].getLastPathComponent());
                if (ctr < length + 1)
                    buffer.append('\n');
            }

            try
            {
                Clipboard clipboard = TOOLKIT.getSystemClipboard();

                StringSelection selection = new StringSelection(
                    buffer.toString()
                    );

                clipboard.setContents(selection, selection);
            }
            catch (Exception e)
            {

            }
        }
    }

    /**
     * Returns wheter or not the paste operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to paste() should be permitted.
     */
    public boolean canPaste()
    {
        return false;
    }

    /**
     * Inserts the contents of the clipbaord into the object.  <b>Optional</b>: Before inserting
     * the contents of the clipboard, remove the current selection from the object.
     */
    public void paste()
    {

    }

    /**
     * Returns wheter or not the selectAll operation can be performed on this ClipboardTarget.
     *
     * @return True iff a call to selectAll() should be permitted.
     */
    public boolean canSelectAll()
    {
        return isEnabled() &&
               getSelectionModel().getSelectionMode() != TreeSelectionModel.SINGLE_TREE_SELECTION;
    }

    /**
     * Selects everything in the object.
     */
    public void selectAll()
    {
        if (isEnabled() &&
            getSelectionModel().getSelectionMode() != TreeSelectionModel.SINGLE_TREE_SELECTION)
        {
            setSelectionInterval(0, getRowCount() - 1);
        }
    }

    /**
     * Intercept FocusEvents to notify the ClipboardTargetManager of the event.
     */
    protected void processFocusEvent(FocusEvent e)
    {
        if (e.getID() == FocusEvent.FOCUS_GAINED)
        {
            ClipboardTargetManager.targetGainedFocus(this);
        }
        else if (e.getID() == FocusEvent.FOCUS_LOST)
        {
            ClipboardTargetManager.targetLostFocus(this);
        }

        super.processFocusEvent(e);
    }

    /**
     * Intercept MouseEvents to activate the popupMenu at the right time.
     */
    protected void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        boolean passedToSuper = false;

        if (popupMenu != null && e.isPopupTrigger())
        {
            if (!hasFocus())
                requestFocus();

            int x = e.getX();
            int y = e.getY();
            Point componentLocation = getLocationOnScreen();
            Dimension screenSize = TOOLKIT.getScreenSize();
            Dimension preferredSize = popupMenu.getPreferredSize();

            if (componentLocation.x + x + preferredSize.width > screenSize.width)
                x -= preferredSize.width;

            if (componentLocation.y + y + preferredSize.height > screenSize.height)
                y -= preferredSize.height;

            e.consume();
            passedToSuper = true;

            super.processMouseEvent(e);

            popupMenu.show(this, x, y);
        }
        else if (e.getID() == MouseEvent.MOUSE_CLICKED && e.getClickCount() == 2 &&
                 (e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)
        {
            // Dispatch the events
            ActionEvent actionEvent;
            String actionCommand = this.actionCommand;
            if (actionCommand == null)
            {
                Object value = getSelectionPath().getLastPathComponent();
                if (value == null)
                    actionCommand = "";
                else
                    actionCommand = value.toString();
            }
            
            actionEvent = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, actionCommand);

            Object array[] = actionListeners.toArray();
            for (int ctr = 0; ctr < array.length; ctr++)
                ((ActionListener)array[ctr]).actionPerformed(actionEvent);
        }

        if (!passedToSuper)
            super.processMouseEvent(e);
    }

    /**
     * Override the setEnabled() method to intercept any changes concerning the enabled status of
     * the component.
     */
    public void setEnabled(boolean enabled)
    {
        super.setEnabled(enabled);

        ClipboardTargetManager.targetAbilityChanged(this);
    }

    /**
     * Override the setSelectionModel method to keep our listener on the correct model.
     */
    public void setSelectionModel(TreeSelectionModel selectionModel)
    {
        if (!(selectionModel instanceof PropertiedListSelectionModel))
            throw new IllegalArgumentException();

        TreeSelectionModel model = getSelectionModel();
        model.removePropertyChangeListener(listener);

        super.setSelectionModel(selectionModel);

        model = getSelectionModel();
        model.addPropertyChangeListener(listener);
    }

    /**
     * Gets the popup menu for this component.
     *
     * @return The popup menu for the component.  Null is returned if this component does not have
     * a popup menu.
     */
    public JPopupMenu getPopupMenu()
    {
        return popupMenu;
    }

    /**
     * Sets the popup menu for this component.
     *
     * @param popupMenu The popup menu for the component.  A null value indicates that this
     * component should not have a popup menu.
     */
    public void setPopupMenu(JPopupMenu popupMenu)
    {
        if (popupMenu != this.popupMenu)
        {
            if (popupMenu != null && this.popupMenu.isVisible())
            {
                this.popupMenu.setVisible(false);
            }

            JPopupMenu oldValue = this.popupMenu;
            this.popupMenu = popupMenu;

            firePropertyChange("popupMenu", oldValue, popupMenu);
        }
    }

    /**
     * The Listener listenes to the selection model of this list to capture PropertyChangeEvents.
     *
     * @author btsang
     * @version 7.1
     */
    protected class Listener implements PropertyChangeListener
    {
        /**
         * Creates a new Listener.
         */
        protected Listener()
        {

        }

        /**
         * Implement PropertyChangeListener to notify the ClipboardTargetManager of a possible
         * change in the canX() methods.
         */
        public void propertyChange(PropertyChangeEvent e)
        {
            if ("selectionMode".equals(e.getPropertyName()))
                ClipboardTargetManager.targetAbilityChanged(PTree.this);
        }
    }
}