package edu.princeton.toy;

import java.awt.*;
import java.awt.event.*;

import edu.princeton.swing.*;
import edu.princeton.swing.text.*;
import edu.princeton.toy.lang.*;

/**
 * TCommentingAutoCompleter is an subclass of TUncommentingAutoCompleter which will also
 * automatically update or add an autocomment when the user completes or modifies a command.
 *
 * @version 7.1
 * @author btsang
 */

public class TCommentingAutoCompleter extends TUncommentingAutoCompleter
{
    public TCommentingAutoCompleter()
    {

    }

    /**
     * The PHighlightedTextArea will pass all key events to its auto-completer before processing
     * the event for itself.
     *
     * @param comp The component which recieved the KeyEvent.
     * @param e The event which occurred.
     * @return True if the component should ignore the KeyEvent (because this method already
     * performed a special action associated with it).  In either case, KeyListeners of the
     * component will still be informed of the event.
     */
    public boolean interceptKeyEvent(PHighlightedTextArea comp, KeyEvent e)
    {
        // Allow the indenter to try to intercept the KeyEvent
        if (super.interceptKeyEvent(comp, e))
            return true;

        // Check for the correct keystroke
        if (e.getID() == KeyEvent.KEY_TYPED && TWord.isHexDigit(e.getKeyChar()))
        {
            int selectionDot = comp.getSelectionDot();

            // Make sure we're in insert mode
            if (!comp.getInsertMode())
                return false;

            // Make sure nothing is selected
            if (selectionDot != comp.getSelectionMark())
                return false;

            HighlightedDocument document = comp.getDocument();
            Point coordinate = document.offsetToCoordinate(selectionDot);

            // Make sure the caret is in the right location
            if (coordinate.x != 7)
                return false;

            String line = document.getText(
                selectionDot - 7,
                document.coordinateToOffset(
                    Integer.MAX_VALUE,
                    coordinate.y
                    )
                );

            // Make sure that the text before the caret is almost a command
            if (!TWord.isHexDigit(line.charAt(0)) || !TWord.isHexDigit(line.charAt(1)) ||
                line.charAt(2) != ':' || line.charAt(3) != ' ' ||
                !TWord.isHexDigit(line.charAt(4)) || !TWord.isHexDigit(line.charAt(5)) ||
                !TWord.isHexDigit(line.charAt(6)))
                return false;

            // Make sure that there is nothing of consequence following the caret
            int length = line.length();
            for (int ctr = 7; ctr < length; ctr++)
                if (line.charAt(ctr) != ' ')
                    return false;

            // Insert the character
            document.insertString(
                selectionDot,
                String.valueOf(e.getKeyChar()),
                true
                );

            // Insert the autocomment
            try
            {
                TWord command = TWord.parseWord(line.substring(4, 7) + e.getKeyChar(), 16);
                int arrayLength = TProgramDocument.COMMENT_COLUMN - 8;
                char array[] = new char[arrayLength];

                array[0] = ' ';
                array[1] = ' ';
                array[2] = ' ';

                String pseudoCodeString = command.toPseudoCodeString(false);
                int pseudoCodeStringLength = pseudoCodeString.length();
                pseudoCodeString.getChars(0, pseudoCodeStringLength, array, 3);

                for (int ctr = pseudoCodeStringLength + 3; ctr < arrayLength; ctr++)
                    array[ctr] = ' ';

                document.replace(
                    selectionDot + 1,
                    length - 7,
                    new String(array),
                    false
                    );
            }
            catch (NumberFormatException err)
            {
                err.printStackTrace();
            }

            return true;
        }

        return false;
    }
}
